from subs2cia.sources import Stream
import pycountry
import logging


def picker(streams: [Stream], target_lang: str = None, forced_stream: int = None):
    r"""
    Returns streams by priority. Streams which are not part of a container are preferred first,
    followed by manually specified stream indices, then streams which match a specified language, finally followed
    by the remaining streams.
    :param streams: List of Stream objects
    :param target_lang: Target language
    :param forced_stream:
    :return:
    """
    for s in streams:
        if s.is_standalone():
            yield s

    if forced_stream is not None:
        yield streams[forced_stream]

    if target_lang is not None:
        target_lang = pycountry.languages.lookup(target_lang)
        target_lang = target_lang.alpha_3

    for s in streams:
        if s.is_standalone():
            continue
        if s.get_language() == target_lang:
            yield s

    for s in streams:
        if s.is_standalone():
            continue
        if s.get_language == target_lang:
            continue
        yield s



# each output has a set of possible input files
# all source_files does is take all of the input files and partitons them into three lists:

from pathlib import Path
import logging
import ffmpeg
from collections import defaultdict
from subs2cia.ffmpeg_tools import ffmpeg_demux
import pycountry


class AVSFile:
    def __init__(self, filepath: Path):
        if not filepath.exists():
            raise AssertionError(f"File {filepath} does not exist")
        self.filepath = filepath
        self.info = None
        self.type = None

    # returns string-encoded type (subtitle, audio, video)
    # determining type may just be as simple as reading the extension
    # but sometimes its better to use a parser and make sure the extension is correct
    def probe(self):
        try:
            self.info = ffmpeg.probe(self.filepath, cmd='ffprobe')
        except ffmpeg.Error as e:
            logging.warning(
                f"Couldn't probe file, skipping {str(self.filepath)}. ffmpeg output: \n" + e.stderr.decode("utf-8"))

        logging.debug(f"ffmpeg probe results: {self.info}")

    def get_type(self):
        if self.info is None:
            self.type = 'unknown'
            return
        if 'streams' not in self.info:
            logging.warning(f"Unexpected ffmpeg.probe output, ignoring file {str(self.filepath)}")
            logging.debug(self.info)
            self.type = None
            return
        if len(self.info['streams']) > 1:
            self.type = 'video'  # video files are treated as multi-stream objects
            return
        stream = self.info['streams'][0]
        if stream['codec_type'] == 'video':
            logging.info(f"File {str(self.filepath)} contains no audio or subtitle tracks!")
        self.type = stream['codec_type']

    def __repr__(self):
        return str(self.filepath)


# single ffmpeg stream
class Stream:
    index = None

    def __init__(self, file: AVSFile, type, stream_info, index=None, ):
        self.file = file
        self.index = index
        self.type = type
        # index of None indicates that demuxing with ffmpeg is not nessecary to extract data
        self.demux_file = None
        self.lang = 'unknownlang'
        self.stream_info = stream_info

        self.get_language()

    def is_standalone(self):
        if self.index is None:
            return True
        return False

    def __repr__(self):
        if self.is_standalone():
            return str(self.file)
        else:
            return f"stream {self.index} ({self.type}) in {self.file}"

    def get_language(self):
        if self.lang != 'unknownlang':
            return self.lang.alpha_3
        if self.is_standalone():
            # no metadata to analyze, look in suffixes for language codes
            suffixes = self.file.filepath.suffixes
            if len(suffixes) >= 2:
                if suffixes[-2] != 'forced':
                    self.lang = suffixes[-2]
                elif len(suffixes) != 2:
                    self.lang = suffixes[-3]
            if self.lang == 'unknownlang':
                return self.lang
            try:
                lang = pycountry.languages.lookup(self.lang)
            except:
                logging.warning(f"{self.lang} is not a language, treating {self.file.filepath} as unknown language")
                self.lang = 'unknownlang'
                return self.lang
            self.lang = lang
            return self.lang.alpha_3
        # look at metadata for language codes
        if 'tags' not in self.file.info['streams'][self.index]:
            return self.lang
        if 'language' not in self.file.info['streams'][self.index]['tags']:
            return self.lang
        try:
            self.lang = pycountry.languages.lookup(self.file.info['streams'][self.index]['tags']['language'])
        except LookupError as e:
            logging.warning(f"{self} language {self.file.info['streams'][self.index]['tags']['language']} is not a "
                            f"proper language code, setting to unknown language.")
            self.lang = 'unknownlang'
            return self.lang
        return self.lang.alpha_3

    def demux(self, overwrite_existing: bool):
        if self.demux_file is not None:
            return self.demux_file
        demux_path = self.file.filepath
        if not self.is_standalone():
            if self.type == 'subtitle':
                subtitle_mapping = {
                    'subrip': 'srt',
                    'ass': 'ass'
                }
                # todo: bitmap subtitles
                if self.stream_info is not None and 'codec_name' in self.stream_info:
                    if self.stream_info['codec_name'] not in subtitle_mapping:
                        extension = 'ass'
                        logging.warning(f"Unknown subtitle type {self.stream_info['codec_name']} found, "
                                     f"will attempt to convert to .ass")
                    else:
                        extension = subtitle_mapping[self.stream_info['codec_name']]

            if self.type == 'audio':
                # we could change what type to demux as similarly to subtitles,
                # but it may cause compatability issues down the road so let's
                # keep it as flac for now
                extension = 'flac'
            demux_path = self.file.filepath.parent / Path(
                f'{self.file.filepath.name}.stream{self.index}.{self.type}.{self.get_language()}.{extension}')

            if overwrite_existing or not demux_path.exists() or demux_path.stat().st_size == 0:
                demux_path = ffmpeg_demux(self.file.filepath, self.index, demux_path)
                if demux_path is None:
                    logging.error(
                        f"Couldn't demux stream {self.index} from {str(self.file.filepath)} (type={self.type})")
                    return
        self.demux_file = AVSFile(demux_path)
        self.demux_file.probe()
        self.demux_file.get_type()
        return self.demux_file

    def cleanup_demux(self):
        if self.demux_file is not None and self.index is not None:
            logging.info(f"Deleting temporary file {str(self.demux_file.filepath)}")
            self.demux_file.filepath.unlink()

    # return a readable path to the data
    def get_data_path(self) -> Path:
        if self.is_standalone() or self.type == 'video':
            return self.file.filepath
        else:
            return self.demux_file.filepath


def common_count(t0, t1):
    # returns the length of the longest common prefix
    i = 0
    for i, pair in enumerate(zip(t0, t1)):
        if pair[0] != pair[1]:
            return i
    return i + 1


# todo: needs work for batching to work
def group_by_longest_prefix(sources: [AVSFile]):
    files = sources
    out = []
    longest = 0
    for f in files:
        # splits = str(f.filepath.name).split('.')
        splits = str(f.filepath.name)
        if out:
            # common = common_count(splits, str(out[-1].filepath.name).split('.'))
            common = common_count(splits, str(out[-1].filepath.name))
            if common < longest:
                yield out
                longest = 0
                out = []
                # otherwise, just update the target prefix length
            else:
                longest = common

                # add the current entry to the group
        out.append(f)

    # return remaining entries as the last group
    if out:
        yield out


def is_language(s):
    try:
        pycountry.languages.lookup(s)
        return True
    except:
        # print(f"{s} is not a language")
        return False


def strip_extensions(p: Path) -> Path:
    if len(p.suffixes) <= 1:
        return p.with_suffix('')
    p = p.with_suffix('')
    if len(p.suffixes) > 1:
        if p.suffixes[-1] == '.forced':
            p = p.with_suffix('')
    if len(p.suffixes) >= 1:
        lcode = p.suffixes[-1]
        lcode = lcode.replace('.', '')
        if is_language(lcode):
            return p.with_suffix('')
        return p


# todo: needs work for batching to work
def group_names(sources: [AVSFile]):
    files = sources
    out = []
    longest = 0
    for f in files:
        # splits = str(f.filepath.name).split('.')
        splits = str(strip_extensions(f.filepath).name)
        if out:
            # common = common_count(splits, str(out[-1].filepath.name).split('.'))
            common = common_count(splits, str(strip_extensions(out[0].filepath).name))
            # if common < longest:
            #     yield out
            #     longest = 0
            #     out = []
            #     # otherwise, just update the target prefix length
            # else:
            #     longest = common
            #     # add the current entry to the group
            # print(splits[0:common])
            # print(strip_extensions(out[0].filepath).name)
            if common == 0 or \
                    strip_extensions(out[0].filepath).name != splits[0:common] or \
                    (len(splits) > common and splits[common] != '.'):
                # no similarity, or similarity does not contain the entire previous name
                yield out
                longest = 0
                out = []
            else:
                longest = common
        out.append(f)

    # return remaining entries as the last group
    if out:
        yield out


def group_files(sources: [AVSFile]):
    file_groups = list(group_names(sources))
    logging.debug(f"groups: {[[f.filepath for f in g] for g in file_groups]}")
    return file_groups



import argparse
from pathlib import Path
import sys


def get_args_subzipper():
    parser = argparse.ArgumentParser(description=f'SubZipper: Map video files to subtitle files')

    parser.add_argument('-s', '--subtitle', metavar='<input files>', dest='subfiles', default=None, required=True,
                        type=str, nargs='+',
                        help='List of subtitle files. Number of subtitle files should equal number of reference files.')

    parser.add_argument('-r', '--reference', metavar='<input files>', dest='reffiles', default=None, required=True,
                        type=str, nargs='+',
                        help='List of reference files, typically video files. '
                             'Number of subtitle files should equal number of reference files. ')

    parser.add_argument('-l', '--language', metavar='ISO_LANG_CODE', dest='lang', default=None, required=False,
                        type=str,
                        help='Language code to append to end of subtitle file. Optional. '
                             'If set, will be checked for validity.')

    parser.add_argument('-ns', '--no-sort', action='store_true', dest='no_sort', default=False,
                        help="If set, will not sort input files alphabetically.")

    parser.add_argument('-d', '--dry-run', action='store_true', dest='dry_run', default=False,
                        help="If set, will print out mappings but will not write any changes to disk.")

    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose', default=False,
                        help='Verbose output if set.')

    # parser.add_argument('-vv', '--debug', action='store_true', dest='debug', default=False,
    #                     help='Verbose and debug output if set')

    args = parser.parse_args()
    return args


def get_args_subs2cia():
    parser = argparse.ArgumentParser(description=f'subs2cia: Extract subtitled dialogue from audiovisual media for use '
                                                 f'in language acquisition ')

    # todo: implement directory parsing
    parser.add_argument('-i', '--inputs', metavar='<input files>', dest='infiles', default=None, required=False,
                        type=str, nargs='+',
                        help='Paths to input files or a single path to a directory of input files.')


    parser.add_argument('-si', '--subtitle-index',  metavar='<index>', dest='subtitle_stream_index', default=None,
                        type=int,
                        help='Force a certain subtitle stream to use. Takes precedence over --target-language option.'
                             'If any input files are standalone subtitle files, they will be used first. '
                             'Use --list-streams for a list of available streams and their indices.')

    parser.add_argument('-ai', '--audio-index',  metavar='<index>', dest='audio_stream_index', default=None,
                        type=int,
                        help='Force a certain subtitle audio to use. Takes precedence over --target-language option.'
                             'If any input files are standalone audio files, they will be used first. '
                             'Use --list-streams for a list of available streams and their indices.')

    parser.add_argument('-b', '--batch', action='store_true', dest='batch', default=False,
                        help='If set, attempts to split input files into groups, one output file per group. '
                             'Groups are determined by file names. If two files share the same root name, such as '
                             '"video0.mkv" and "video0.srt", then they are part of the same group.')

    parser.add_argument('-u', '--dry-run', action='store_true', dest='dry_run', default=False,
                        help="If set, will analyze input files but won't demux or generate any condensed files")

    parser.add_argument('-o', '--output-name', metavar='<name>', dest='outstem', default=None,
                        type=str,
                        help='Output file name to save to, without the extension (specify extension using -ae or -ve). '
                             'By default, uses the file name of the first input file with the input extension removed '
                             'and "condensed.{output extension} added. \n'
                             'Ignored if batch mode is enabled.')

    parser.add_argument('-d', '--output-dir', metavar='/path/to/directory', dest='outdir', default=None,
                        type=str,
                        help='Output directory to save to. Default is the directory the input files reside in.')

    parser.add_argument('-ae', '--audio-extension', metavar='<audio extension>', dest='out_audioext', default='mp3',
                        type=str,
                        help='Condensed audio extension to save as (without the dot). '
                             'Default is mp3, flac has been tested to work.')
    # todo: dot stripper, output naming
    parser.add_argument('-m', '--gen-video', action='store_true', dest='condensed_video', default=False,
                        help='If set, generates condensed video along with condensed audio and subtitles. '
                             'Subtitles are muxed in to video file. '
                             'WARNING: VERY CPU INTENSIVE).')

    parser.add_argument('--overwrite-on-demux', action='store_true', dest='demux_overwrite_existing', default=False,
                        help='If set, will overwrite existing files when demuxing temporary files.')

    parser.add_argument('--keep-temporaries', action='store_true', dest='keep_temporaries', default=False,
                        help='If set, will not delete any demuxed temporary files.')

    parser.add_argument('--no-overwrite-on-generation', action='store_false', dest='overwrite_existing_generated',
                        default=True,
                        help='If set, will not overwrite existing files when generating condensed media.')

    parser.add_argument('-ni', '--ignore-none', action='store_true', dest='use_all_subs',
                        default=False,
                        help='If set, will not use internal heuristics to remove non-dialogue lines from the subtitle. '
                             'Overridden by -R.')

    parser.add_argument('-R', '--sref', metavar='<regular expression>', dest='subtitle_regex_filter', default=None,
                        type=str,
                        help='For filtering non-dialogue subtitles. Lines that match given regex are IGNORED '
                             'during subtitle processing and will not influence condensed audio. '
                             'Ignored lines may be included in condensed subtitles. This option will override the '
                             'internal subs2cia non-dialogue filter.')

    parser.add_argument('-p', '--padding', metavar='msecs', dest='padding', default=0,
                        type=int,
                        help='Adds this many milliseconds of audio before and after every subtitle. '
                             'Overlaps with adjacent subtitles are merged automatically.')

    parser.add_argument('-t', '--threshold', metavar='msecs', dest='threshold', default=0,
                        type=int,
                        help="If two subtitles start and end within (threshold + 2*padding) "
                             "milliseconds of each other, they will be merged. Useful for preserving silences between "
                             "subtitle lines.")

    parser.add_argument('-r', '--partition', metavar='secs', dest='partition', default=0,
                        type=int,
                        help="If set, attempts to partition the input audio into "
                             "seperate blocks of this size seconds BEFORE condensing. Partitions and splits respect "
                             "subtitle boundaries and will not split a single subtitle across two output files."
                             " 0 partition length is ignored. For example, if the partition size is 60 seconds and the "
                             "input media is 180 seconds long, then there will be three output files. The first output "
                             "file will contain condensed media from the first 60 seconds of the source material, "
                             "the second output file will contain the next 60 seconds of input media, and so on.")

    parser.add_argument('-s', '--split', metavar='secs', dest='split', default=0,
                        type=int,
                        help="If set, attempts to split the condensed audio into "
                             "seperate blocks of this size AFTER condensing. 0 "
                             "split length is ignored. "
                             "Partitions and splits respect subtitle boundaries and will not split a single subtitle "
                             "across two output files. "
                             "Done within a partition. For example, say the split length is "
                             "60 seconds and the condensed audio length of a input partition is 150 seconds. "
                             "The output file will be split into three files, the first two ~60 seconds long and the "
                             "last ~30 seconds long.")

    parser.add_argument('-c', '--minimum-compression-ratio', metavar='<ratio>', dest='minimum_compression_ratio',
                        default=0.2, type=float,
                        help="Will only generate from subtitle files that are this fraction long of the selected audio "
                             "file. Default is 0.2, meaning the output condensed file must be at least 20% as long as "
                             "the chosen audio stream. If the output doesn't reach this minimum, then a different "
                             "subtitle file will be chosen, if available. Used to ignore subtitles that contain only"
                             "signs and songs.")

    parser.add_argument('-tl', '--target-language', metavar='ISO_code', dest='target_lang', default=None,
                        type=str,
                        help='If set, attempts to use audio and subtitle files that are in this language first. '
                             'Should follow ISO language codes. ')

    # todo: consider depreciating this option
    parser.add_argument('-a', '--absolute-paths', action='store_true', dest='absolute_paths', default=False,
                        help='Prints absolute paths from the root directory instead of given paths.')

    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose', default=False,
                        help='Verbose output if set.')

    parser.add_argument('-vv', '--debug', action='store_true', dest='debug', default=False,
                        help='Verbose and debug output if set')

    parser.add_argument('--preset', metavar='preset#', dest='preset', type=int, default=None,
                        help='If set, uses a given preset. User arguments will override presets.')
    parser.add_argument('-lp', '--list-presets', dest='list_presets', action='store_true', default=False,
                        help='Lists all available built-in presets and exits.')
    parser.add_argument('-ls', '--list-streams', dest='list_streams', action='store_true', default=False,
                        help='Lists all audio, subtitle, and video streams found in given input files and exits.')
    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)

    return args



from subs2cia.sources import Stream
from subs2cia.ffmpeg_tools import ffmpeg_demux

import logging
import pysubs2 as ps2  # for reading in subtitles
from pathlib import Path
from datetime import timedelta
import ffmpeg
import re
import copy


class SubGroup:
    def __init__(self, events: [ps2.SSAEvent], ephemeral: bool, threshold: int, padding: int):
        self.contains_only_ephemeral = ephemeral  # if true, won't affect merging/splitting
        self.events = events
        self.ephemeral_events = []  # not empty only when mixing ephemeral and dialogue events

        self.threshold = threshold
        self.padding = padding

    @property
    def events_start(self):
        lowest = float('inf')
        for e in self.events:  # there may be 0 events
            if e.start < lowest:
                lowest = e.start
        return lowest

    @property
    def events_end(self):
        highest = 0
        for e in self.events:  # there may be 0 events
            if e.end > highest:
                highest = e.end
        return highest

    @property
    def group_range(self):
        if self.contains_only_ephemeral:
            return [self.events_start, self.events_end]
        grange = self.padding
        return [self.events_start - grange if self.events_start - grange > 0 else 0, self.events_end + grange]

    @property
    def group_limits(self):
        if self.contains_only_ephemeral:
            return [self.events_start, self.events_end]
        limit = self.threshold/2 + self.padding  # divide by two: threshold is distance to next group
        return [self.events_start - limit if self.events_start - limit > 0 else 0, self.events_end + limit]

    def __repr__(self):
        s = f"<SubGroup, |{self.group_limits[0]} {self.group_range[0]} {(self.events_start, self.events_end)} {self.group_range[1]} {self.group_limits[1]}|]"
        return s


# TODO
class SubtitleManipulator:
    def __init__(self, subpath: Path, threshold: int, padding: int):
        self.subpath = subpath
        self.ssadata = None
        self.condensed_ssadata = None
        self.ssa_events = None
        self.groups = None
        self.ephemeral = None

        self.threshold = threshold
        self.padding = padding

    def load(self, include_all, regex):
        if not self.subpath.exists():
            logging.warning(f"Subtitle file {self.subpath} does not exist")
            return

        logging.debug(f"Loading subtitles at {self.subpath}")
        try:
            self.ssadata = ps2.load(str(self.subpath))
        except (ValueError, AttributeError) as e:
            logger = logging.getLogger(__name__)
            logger.exception(e)
            logging.warning(f"pysub2 enountered error loading subtitle file {self.subpath}")
            self.ssadata = None
            return
        logging.debug(f"Loaded {self.ssadata}")

        self.ssa_events = self.ssadata.events
        self.ssa_events.sort(key=lambda x: x.start)

        self.groups = []
        for e in self.ssa_events:
            self.groups.append(SubGroup([e], ephemeral=not is_dialogue(e, include_all, regex),
                                        threshold=self.threshold,
                                        padding=self.padding))

    def merge_groups(self):
        merged = []
        self.ephemeral = []
        for group in self.groups:
            # ephermal groups are not merged with any other groups so they can be dealt with seperately
            if group.contains_only_ephemeral:
                self.ephemeral.append(group)
                continue
            if len(merged) == 0:  # first group
                merged.append(group)
                continue
            if merged[-1].group_limits[1] > group.group_limits[0]:
                merged[-1].events += group.events
            else:
                merged.append(group)
        self.groups = merged
        logging.debug(f"Merged groups {merged}")
        # add ephemeral groups back into rest of groups
        for egroup in self.ephemeral:  # assuming each egroup contains one ssaevent
            for group in self.groups:
                # if egroup is valid inside group range
                    # create a new ssaevent from egroup that fits inside group
                if group.group_range[0] < egroup.events_start < group.group_range[1] or \
                group.group_range[0] < egroup.events_end < group.group_range[1]:
                    new_event = egroup.events[0].copy()
                    new_event.start = new_event.start if new_event.start > group.group_range[0] else group.group_range[0]
                    new_event.end = new_event.end if new_event.end < group.group_range[1] else group.group_range[1]
                    group.ephemeral_events.append(new_event)
        logging.debug("Inserted ephemeral events")

    def get_times(self):
        # returns list of tuples of group ranges
        times = []
        for g in self.groups:
            times.append(g.group_range)
        return times

    def condense(self):
        r"""
        laststart = 0
        for each group g
            shift all event times in g back by g.group_range[0] - laststart milliseconds
        :return:
        """
        laststart = 0
        groups = copy.deepcopy(self.groups)
        for g in groups:
            shift = g.group_range[0] - laststart
            for e in g.events:
                e.start -= shift
                e.end -= shift
            for e in g.ephemeral_events:
                e.start -= shift
                e.end -= shift
            laststart = g.group_range[1]
        logging.debug("Shifted subtitle groups")
        # extract shifted SSAevents
        condensed_events = []
        for g in groups:
            for e in g.events:
                condensed_events.append(e)
            for e in g.ephemeral_events:
                condensed_events.append(e)
        self.condensed_ssadata = copy.copy(self.ssadata)
        self.condensed_ssadata.events = condensed_events


def decide_partitions(sub_times, partition=0):
    if partition == 0:
        return [(0, len(sub_times)), ]

    partitions = list()
    current_partition = 0
    start = 0
    idx = 0
    for idx, t in enumerate(sub_times):
        current_partition_boundary = (current_partition + 1) * partition
        if t[1] > current_partition_boundary:
            if abs(t[1] - current_partition_boundary) < abs(sub_times[idx - 1][1] - current_partition_boundary):
                end = idx + 1
            else:
                end = idx
            partitions.append((start, end))
            current_partition += 1
            start = end
    end = idx + 1
    partitions.append((start, end))
    return partitions


def split_times(sub_times, partition_indicies, splitsize=0):
    partition_times = sub_times[partition_indicies[0]:partition_indicies[1]]
    if splitsize == 0:
        return [partition_times, ]

    prev = 0
    for idx, time in enumerate(sub_times):
        time.append(time[1] - time[0] + prev)
        prev = time[2]

    start = 0
    idx = 0
    splits = list()
    current_split = 0
    for idx, t in enumerate(partition_times):
        current_partition_boundary = (current_split + 1) * splitsize
        if t[2] > current_partition_boundary:
            if abs(t[2] - current_partition_boundary) < abs(partition_times[idx - 1][2] - current_partition_boundary):
                end = idx + 1
            else:
                end = idx
            splits.append((start, end))
            current_split += 1
            start = end
    end = idx + 1
    splits.append((start, end))
    splittimes = list()
    for split in splits:
        tmp = partition_times[split[0]:split[1]]
        tmp = [[x[0], x[1]] for x in tmp]
        splittimes.append(tmp)
    return splittimes


def partition_and_split(sub_times, partition_size=0, split_size=0):
    # returns a list tuple index pairs in [start, end) format for sub_times
    partitions = decide_partitions(sub_times,
                                   partition=partition_size)
    """
    partition1
        split1
            time1
                start
                end
            time2
                ...
        split2
            time1
        split3
        split4...
    partition2
        split1
            time1
            time2...
    """
    divided_times = list()
    for p, partition_size in enumerate(partitions):
        # returns list of list of times in milliseconds, can feed into condense_audio
        times = split_times(sub_times, partition_size,
                            splitsize=split_size)
        divided_times.append(times)

    return divided_times


# given raw subtitle timing data, merge overlaps and perform padding and merging
# DEPRECIATED
def merge_times(times: [], threshold=0, padding=0):
    # padding: time to add to the beginning and end of each subtitle. Resulting overlaps are removed.
    # threshold: if two subtitles are within this distance apart, they will be merged in the audio track
    times.sort(key=lambda x: x[0])  # sort by first value in tuple

    initial = len(times)
    if padding > 0:
        for time in times:
            time[0] -= padding
            time[1] += padding

    # even if threshold is 0 any added padding may cause overlaps, which the code below will also merge
    idx = 0
    while idx < len(times) - 1:
        if times[idx + 1][0] < times[idx][1] + threshold:
            # if the next subtitle starts within 3 seconds of the current subtitle finishing, merge them
            times[idx][1] = times[idx + 1][1]
            del times[idx + 1]
            if idx < len(times) - 1:
                continue
        idx += 1

    logging.info(f"Merged {initial} subtitles into {len(times)} audio snippets\n")
    logging.debug(f"Merged times: {times}")
    return times


# examine a subtitle's text and determine if it contains text or just signs/song styling
# could be much more robust, by including regexes
def is_dialogue(line, include_all=False, regex=None):
    if regex is not None:
        p = re.compile(regex)
        logging.debug(f"using regex {regex} for subtitle search")
        m = p.findall(line.text)
        if len(m) >= 1:
            logging.debug(f'Regex {regex} found match(s) in "{line.text}": {m}')
            return False
        return True
    if include_all:  # ignore filtering
        return True
    if line.type != "Dialogue":
        return False
    if len(line.text) == 0:
        return False
    if '{' == line.text[0]:
        return False
    if '（' == line.text[0] and '）' == line.text[-1]:
        return False

    # list of characters that, if present anywhere in the subtitle text, means that the subtitle is not dialogue
    globally_invalid = ["♪", "♪"]
    if any(invalid in line.text for invalid in globally_invalid):
        return False
    return True


# given a path to a subtitle file, load it in and strip it of non-dialogue text
# keyword arguments are filter options that are passed to is_spoken_dialogue for filtering configuration
# DEPRECIATED
def load_subtitle_times(subfile: Path, include_all_lines=False):
    # if dry_run:
    #     print("dry_run: will load", subfile)
    #     if os.path.isfile(subfile) is False:
    #         print(f"note: {subfile} doesn't exist yet, skipping subtitle loading")
    #         return [[0, 0], ]  # might just be a demuxed file that hasn't been demuxed yet
    logging.info(f"loading {subfile}")
    try:
        subs = ps2.load(str(subfile))
    except (ValueError, AttributeError) as e:
        logger = logging.getLogger(__name__)
        logger.exception(e)
        logging.warning(f"pysub2 enountered error loading subtitle file {subfile}, trying a different subtitle file...")
        return None
    logging.info(subs)

    times = list()
    count = 0
    for idx, line in enumerate(subs):
        # print(line)
        if is_dialogue(line, include_all=include_all_lines):
            # print(line)
            times.append([line.start, line.end])
        else:
            logging.debug(f"ignoring {line.text}, probably not spoken dialogue")
            count += 1
            pass
    logging.info(f"ignored {count} lines")  # number of subtitles that looked like they didn't contain dialogue

    # print(times)
    if len(times) == 0:
        logging.warning(f"warning: got 0 dialogue lines from subtitle file {subfile}")
        return None
    return times


def get_partitioned_and_split_times_duration(sub_times):
    r"""
    Calculates the total amount of condensed audio to be generated from the subtitle times, in milliseconds
    :param sub_times:
    :return: time in milliseconds
    """
    total = list()
    for split in sub_times:
        for partition in split:
            for x1, x2 in partition:
                total.append(x2 - x1)
    subs_total = sum(total)
    return subs_total


# don't use: not reliable
# def get_audiostream_duration(stream_info):
#     sps = int(
#         stream_info['time_base'].split('/')[1])  # audio samples per second, inverse of sampling frequency
#     samples = stream_info['duration_ts']  # total samples in audio track
#     audio_total = samples / sps * 1000
#     return audio_total


def get_audiofile_duration(audiofile: Path):
    audio_info = ffmpeg.probe(str(audiofile), cmd='ffprobe')
    sps = int(
        audio_info['streams'][0]['time_base'].split('/')[1])  # audio samples per second, inverse of sampling frequency
    samples = audio_info['streams'][0]['duration_ts']  # total samples in audio track
    audio_total = samples / sps * 1000
    return audio_total


def get_compression_ratio(sub_times, audiofile: Path, verbose=True):
    # audio_info = ffmpeg.probe(str(audiofile), cmd='ffprobe')
    # sps = int(
    #     audio_info['streams'][0]['time_base'].split('/')[1])  # audio samples per second, inverse of sampling freq
    # samples = audio_info['streams'][0]['duration_ts']  # total samples in audio track
    # audio_total = samples / sps * 1000
    audio_total = get_audiofile_duration(audiofile)
    # total = list()
    # for split in sub_times:
    #     for partition in split:
    #         for x1, x2 in partition:
    #             total.append(x2-x1)
    # subs_total = sum(total)
    subs_total = get_partitioned_and_split_times_duration(sub_times)
    if verbose:
        logging.info(f"will condense {str(timedelta(milliseconds=audio_total)).split('.')[0]} of source audio to "
                     f"{str(timedelta(milliseconds=subs_total)).split('.')[0]} "
                     f"({round(subs_total / audio_total * 100, 1)}% compression ratio) of condensed audio")
    return subs_total / audio_total



import ffmpeg
import logging
from pathlib import Path
import os
import tempfile
import subprocess


# given a stream in the input file, demux the stream and save it into the outfile with some type
def ffmpeg_demux(infile: Path, stream_idx: int, outfile: Path):
    # output format is specified via extention on outfile
    logging.debug(f"demuxing stream {stream_idx} from file {infile} to {outfile}")
    video = ffmpeg.input(infile)
    stream = video[str(stream_idx)]  # don't need 0
    stream = ffmpeg.output(stream, str(outfile))
    stream = ffmpeg.overwrite_output(stream)
    logging.debug(f"ffmpeg arguments: {ffmpeg.get_args(stream)}")
    try:
        ffmpeg.run(stream, quiet=logging.getLogger().getEffectiveLevel() >= logging.WARNING)  # verbose only
    except ffmpeg.Error as e:
        if e.stderr is None:
            logging.warning(
                f"Couldn't demux stream {stream_idx} from {infile}, skipping.")
            return None
        logging.warning(
            f"Couldn't demux stream {stream_idx} from {infile}, skipping. ffmpeg output: \n" + e.stderr.decode("utf-8"))
        return None
    return outfile


# from ffmpeg-python _run.py
class Error(Exception):
    def __init__(self, cmd, stdout, stderr):
        super(Error, self).__init__(
            '{} error (see stderr output for detail)'.format(cmd)
        )
        self.stdout = stdout
        self.stderr = stderr


def ffmpeg_condense_audio(audiofile, sub_times, outfile=None):
    if outfile is None:
        outfile = "condensed.flac"
    # logging.info(f"saving condensed audio to {outfile}")

    # get samples in audio file
    audio_info = ffmpeg.probe(audiofile, cmd='ffprobe')
    sps = int(
        audio_info['streams'][0]['time_base'].split('/')[1])  # audio samples per second, inverse of sampling frequency
    # samples = audio_info['streams'][0]['duration_ts']  # total samples in audio track

    stream = ffmpeg.input(audiofile)
    clips = list()
    for time in sub_times:  # times are in milliseconds
        start = int(time[0] * sps / 1000)  # convert to sample index
        end = int(time[1] * sps / 1000)
        # use start_pts for sample/millisecond level precision
        clips.append(stream.audio.filter('atrim', start_pts=start, end_pts=end).filter('asetpts', 'PTS-STARTPTS'))
    combined = ffmpeg.concat(*clips, a=1, v=0)
    if os.path.splitext(outfile)[1] == ".mp3":
        combined = ffmpeg.output(combined, outfile, audio_bitrate='320k')  # todo: make this user-settable
    else:
        combined = ffmpeg.output(combined, outfile)
    combined = ffmpeg.overwrite_output(combined)
    logging.debug(f"ffmpeg arguments: {' '.join(ffmpeg.get_args(combined))}")
    args = ffmpeg.get_args(combined)
    if len("ffmpeg " + " ".join(args)) > 32766 and os.name == 'nt':
        logging.warning("Arguments passed to ffmpeg exceeds 32767 characters while running on a Windows system. "
                        "Will try using a temporary file to pass filter_complex arguments to ffmpeg.")
        idx = args.index("-filter_complex") + 1
        complex_filter = str(args[idx])
        # write complex_filter to a temporary file
        fp = tempfile.NamedTemporaryFile(
            delete=False)  # don't delete b/c can't open file again when it's already open in windows
        fp.write(complex_filter.encode(encoding="utf-8"))
        fp.close()
        args[idx] = fp.name
        args[idx - 1] = "-filter_complex_script"
    args = ["ffmpeg"] + args

    # ffmpeg.run(combined, quiet=logging.getLogger().getEffectiveLevel() >= logging.WARNING)

    pipe_stdin = False
    pipe_stdout = False
    pipe_stderr = False
    quiet = logging.getLogger().getEffectiveLevel() >= logging.WARNING

    stdin_stream = subprocess.PIPE if pipe_stdin else None
    stdout_stream = subprocess.PIPE if pipe_stdout or quiet else None
    stderr_stream = subprocess.PIPE if pipe_stderr or quiet else None
    process = subprocess.Popen(
        args, stdin=stdin_stream, stdout=stdout_stream, stderr=stderr_stream
    )
    out, err = process.communicate(input)
    retcode = process.poll()
    if retcode:
        raise Error('ffmpeg', out, err)


def export_condensed_audio(divided_times, audiofile: Path, outfile=None, use_absolute_numbering=False):
    # outfile is full path with extension
    audiofile = str(audiofile)
    if outfile is not None:
        outfile = str(outfile)

    if outfile is None:  # no output path given, use audiofile path
        outfile = audiofile
    elif outfile[0] == '.' and outfile[1:].isalnum():  # outfile is just an extension, use audiofile for path
        # extension = outfile
        outfile = os.path.splitext(audiofile)[0] + outfile
    else:  # outfile is already full path with extension
        pass
    idx = 0
    for p, partition in enumerate(divided_times):
        if len(partition) == 0:
            continue
        for s, split in enumerate(partition):
            if len(split) == 0:
                continue
            idx += 1
            if use_absolute_numbering:  # todo: remove outfile naming from this function
                outfilesplit = os.path.splitext(outfile)[0] + \
                               f".pt{idx}" + \
                               ".condensed" + \
                               os.path.splitext(outfile)[1]
            else:
                outfilesplit = os.path.splitext(outfile)[0] + \
                               (f".p{p + 1}" if len(divided_times) != 1 else "") + \
                               (f".s{s + 1}" if len(partition) != 1 else "") + \
                               ".condensed" + \
                               os.path.splitext(outfile)[1]

            ffmpeg_condense_audio(audiofile=audiofile, sub_times=split, outfile=outfilesplit)


def export_condensed_video(divided_times, audiofile: Path, subfile: Path, videofile: Path, outfile=None,
                           use_absolute_numbering=False):
    # outfile is full path with extension
    audiofile = str(audiofile)
    if outfile is not None:
        outfile = str(outfile)

    if outfile is None:  # no output path given, use audiofile path
        outfile = audiofile
    elif outfile[0] == '.' and outfile[1:].isalnum():  # outfile is just an extension, use audiofile for path
        # extension = outfile
        outfile = os.path.splitext(audiofile)[0] + outfile
    else:  # outfile is already full path with extension
        pass
    idx = 0
    for p, partition in enumerate(divided_times):
        if len(partition) == 0:
            continue
        for s, split in enumerate(partition):
            if len(split) == 0:
                continue
            idx += 1
            if use_absolute_numbering:
                outfilesplit = os.path.splitext(outfile)[0] + \
                               f".pt{idx}" + \
                               ".condensed" + \
                               os.path.splitext(outfile)[1]
            else:
                outfilesplit = os.path.splitext(outfile)[0] + \
                               (f".p{p + 1}" if len(divided_times) != 1 else "") + \
                               (f".s{s + 1}" if len(partition) != 1 else "") + \
                               ".condensed" + \
                               os.path.splitext(outfile)[1]
            # todo: need to split subfiles with partition, split options
            ffmpeg_condense_video(audiofile=audiofile, videofile=str(videofile), subfile=str(subfile),
                                  sub_times=split, outfile=outfilesplit)


def trim(input_path, output_path, start=30, end=60):
    input_stream = ffmpeg.input(input_path)

    vid = (
        input_stream.video
            .trim(start=start, end=end)
            .setpts('PTS-STARTPTS')
    )
    aud = (
        input_stream.audio
            .filter_('atrim', start=start, end=end)
            .filter_('asetpts', 'PTS-STARTPTS')
    )

    joined = ffmpeg.concat(vid, aud, v=1, a=1).node
    output = ffmpeg.output(joined[0], joined[1], output_path)
    output.run()


def ffmpeg_condense_video(audiofile: str, videofile: str, subfile: str, sub_times, outfile):
    logging.info(f"saving condensed video to {outfile}")

    # get samples in audio file
    audio_info = ffmpeg.probe(audiofile, cmd='ffprobe')
    sps = int(
        audio_info['streams'][0]['time_base'].split('/')[1])  # audio samples per second, inverse of sampling frequency
    # samples = audio_info['streams'][0]['duration_ts']  # total samples in audio track

    audiostream = ffmpeg.input(audiofile)
    videostream = ffmpeg.input(videofile)
    substream = ffmpeg.input(subfile)
    vid = videostream.video.filter_multi_output('split')
    # sub = videostream['s'].filter_multi_output('split')
    aud = audiostream.audio.filter_multi_output('asplit')

    clips = []
    for idx, time in enumerate(sub_times):  # times are in milliseconds
        # start = int(time[0] * sps / 1000)  # convert to sample index
        # end = int(time[1] * sps / 1000)
        start = time[0] / 1000
        end = time[1] / 1000
        # use start_pts for sample/millisecond level precision

        a = aud[idx].filter('atrim', start=start, end=end).filter('asetpts', expr='PTS-STARTPTS')
        v = vid[idx].trim(start=start, end=end).setpts('PTS-STARTPTS')
        # s = sub[idx].trim(start=start, end=end).setpts('PTS-STARTPTS')
        clips.extend((v, a))

    out = ffmpeg.concat(
        *clips,
        v=1,
        a=1
    ).output(substream, outfile)

    # output = ffmpeg.output(joined[0], joined[1], outfile)
    out = ffmpeg.overwrite_output(out)
    logging.debug(f"ffmpeg arguments: {ffmpeg.get_args(out)}")
    ffmpeg.run(out, quiet=logging.getLogger().getEffectiveLevel() >= logging.WARNING)



import setuptools
import pathlib
from subs2cia import __version__

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
long_description = (HERE / "README.md").read_text()

setuptools.setup(
    name="subs2cia",
    version=__version__,
    license='MIT',
    author="Daniel Xing",
    author_email="danielxing97@gmail.com",
    description="A subtitle-based multimedia extractor and compressor",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dxing97/subs2cia",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        "ffmpeg-python",
        "pycountry",
        "pysubs2"
    ],
    entry_points={
        # create a cli command called 'subs2cia' which runs the main() function in subs2cia.cli
        'console_scripts': ['subs2cia=subs2cia.cli:main', 'subzipper=subs2cia.cli:subzipper_main']
    }
)


from subs2cia.sources import AVSFile
from subs2cia.pickers import picker
from subs2cia.sources import Stream
import subs2cia.subtools as subtools
from subs2cia.ffmpeg_tools import export_condensed_audio, export_condensed_video

import logging
from collections import defaultdict
from pathlib import Path


def picked_sources_are_insufficient(d: dict):
    for k in d:
        if d[k] == 'retry':
            return True
    if d['subtitle'] is None:
        return True
    if d['audio'] is None:
        return True
    return False


def insufficient_source_streams(d: dict):
    if len(d['subtitle']) == 0:
        return True
    if len(d['audio']) == 0:
        return True
    return False


class SubCondensed:
    def __init__(self, sources: [AVSFile], outdir: Path, condensed_video: bool, threshold: int, padding: int,
                 partition: int, split: int, demux_overwrite_existing: bool, overwrite_existing_generated: bool,
                 keep_temporaries: bool, target_lang: str, out_audioext: str, minimum_compression_ratio: float,
                 use_all_subs: bool, subtitle_regex_filter: str, audio_stream_index: int, subtitle_stream_index: int):
        r"""

        :param sources: List of AVSFile objects, each representing one input file
        :param outdir: Output directory to save to. Default is the directory of the first source file in *sources*
        :param condensed_video: If set, generates condensed video as well
        :param threshold:
        :param padding:
        :param partition:
        :param split:
        :param demux_overwrite_existing: If set, demuxing operations will overwrite existing files on disk
        :param overwrite_existing_generated: If set, generation operations will overwrite existing files on disk
        :param keep_temporaries: If set, will not delete demuxed files on cleanup
        :param target_lang: Target language for audio AND subtitles
        :param out_audioext: Output audio extension
        :param minimum_compression_ratio: Chosen subtitle stream must yield generated audio at least this percent long of audio file
        """

        if outdir is None:
            self.outdir = sources[0].filepath.parent
        else:
            self.outdir = outdir
        self.outstem = sources[0].filepath.stem
        self.sources = sources
        self.out_audioext = out_audioext
        self.out_vidoeext = '.mkv'
        self.out_subext = None  # extensions must contain dot

        # logging.debug(f'Will save a file with stem "{self.outstem}" to directory "{self.outdir}"')
        logging.info(f"Mapping input file(s) {sources} to one output file")

        self.partitioned_streams = defaultdict(list)

        self.picked_streams = {
            'audio': None,
            'subtitle': None,
            'video': None
        }

        self.pickers = {
            'audio': None,
            'subtitle': None,
            'video': None
        }

        self.target_lang = target_lang
        # following indices overrides target_lang, is overridden by standalone input audio/subtitle files
        self.audio_stream_index = audio_stream_index
        self.subtitle_stream_index = subtitle_stream_index

        self.padding = padding
        self.threshold = threshold
        self.partition = partition
        self.split = split

        self.dialogue_times = None
        self.minimum_compression_ratio = minimum_compression_ratio

        self.demux_overwrite_existing = demux_overwrite_existing
        self.overwrite_existing_generated = overwrite_existing_generated
        self.keep_temporaries = keep_temporaries

        self.condensed_audio = True  # can be exposed later if needed
        self.condensed_video = condensed_video
        self.condensed_subtitles = True  # can be exposed later if needed

        self.use_all_subs = use_all_subs
        self.subtitle_regex_filter = subtitle_regex_filter

        self.insufficient = False
        self.subtitle_outfile = None

    # go through source files and count how many subtitle and audio streams we have
    def get_and_partition_streams(self):
        for sourcefile in self.sources:
            if sourcefile.type == 'video':
                # dig into streams
                for idx, stream_info in enumerate(sourcefile.info['streams']):
                    stype = stream_info['codec_type']
                    self.partitioned_streams[stype].append(Stream(file=sourcefile, type=stype,
                                                                  index=idx, stream_info=stream_info))
                continue
            self.partitioned_streams[sourcefile.type].append(Stream(file=sourcefile, type=sourcefile.type,
                                                                    stream_info=sourcefile.info,
                                                                    index=None))
            # for stream in sourcefile
        for k in self.partitioned_streams:
            logging.info(f"Found {len(self.partitioned_streams[k])} {k} input streams")
            # logging.debug(f"Streams found: {self.partitioned_streams[k]}")

    def initialize_pickers(self):
        for k in self.pickers:
            idx = None
            if k == 'audio':
                idx = self.audio_stream_index
            if k == 'subtitle':
                idx = self.subtitle_stream_index
            self.pickers[k] = picker(self.partitioned_streams[k], target_lang=self.target_lang, forced_stream=idx)

    def list_streams(self):
        print(f"Listing streams found in {self.sources}")
        for k in ['subtitle', 'audio', 'video']:
            print(f"Available {k} streams:")
            for idx, stream in enumerate(self.partitioned_streams[k]):
                desc_str = ''
                if "codec_name" in stream.stream_info:
                    desc_str = desc_str + "codec: " + stream.stream_info['codec_name'] + ", "
                if "tags" in stream.stream_info:
                    tags = stream.stream_info['tags']
                    if "language" in tags:
                        desc_str = desc_str + "lang_code: " + tags['language'] + ", "
                    if "title" in tags:
                        desc_str = desc_str + "title: " + tags['title'] + ", "
                if desc_str == '':
                    desc_str = f"Stream {idx: 3}: no information found"
                else:
                    desc_str = f"Stream {idx: 3}: {desc_str}"
                print(desc_str)
            print("")

    def choose_streams(self):
        if insufficient_source_streams(self.partitioned_streams):
            logging.error(f"Not enough input sources to generate condensed output for output stem {self.outstem}")
            self.insufficient = True
            return
        while picked_sources_are_insufficient(self.picked_streams):
            for k in ['audio', 'video', 'subtitle']:
                if len(self.partitioned_streams[k]) == 0:
                    logging.debug(f"no input streams of type {k}")
                    continue
                if self.picked_streams[k] is None:
                    try:
                        self.picked_streams[k] = next(self.pickers[k])
                    except StopIteration as s:
                        logging.critical("Input streams for this group are invalid for condensing")
                        self.insufficient = True
                        return
            for k in ['audio', 'subtitle', 'video']:
                # validate picked streams

                # todo: spin off into its own function at a later step
                if k == 'audio':
                    afile = self.picked_streams[k].demux(overwrite_existing=self.demux_overwrite_existing)
                    if afile is None:
                        self.picked_streams[k] = None

                if k == 'subtitle':
                    subfile = self.picked_streams[k].demux(
                        overwrite_existing=self.demux_overwrite_existing)  # type AVSFile
                    if subfile is None:
                        self.picked_streams[k] = None
                        continue
                    # times = subtools.load_subtitle_times(subfile.filepath, include_all_lines=self.use_all_subs)
                    subdata = subtools.SubtitleManipulator(subfile.filepath,
                                                           threshold=self.threshold, padding=self.padding)
                    subdata.load(include_all=self.use_all_subs, regex=self.subtitle_regex_filter)
                    if subdata.ssadata is None:
                        self.picked_streams[k] = None
                        continue
                    if self.picked_streams['audio'] is None:
                        # can't verify subtitle validity until audio candidate is found
                        continue
                    # times = subtools.merge_times(times, threshold=self.threshold, padding=self.padding)
                    subdata.merge_groups()
                    times = subdata.get_times()
                    ps_times = subtools.partition_and_split(times, self.partition, self.split)

                    sublength = subtools.get_partitioned_and_split_times_duration(ps_times)
                    audiolength = subtools.get_audiofile_duration(
                        self.picked_streams['audio'].demux_file.filepath)
                    compression_ratio = sublength / audiolength
                    if compression_ratio < self.minimum_compression_ratio:
                        logging.info(f"got compression ratio of {compression_ratio}, which is smaller than the minimum"
                                     f" ratio of {self.minimum_compression_ratio}, retrying wth different subtitle "
                                     f"file. If too many subtitles are being ignored, try -ni or -R. If the minimum "
                                     f"ratio is too high, try -c.")
                        self.picked_streams[k] = None
                        continue
                    self.dialogue_times = subtools.partition_and_split(sub_times=times,
                                                                       partition_size=1000 * self.partition,
                                                                       split_size=1000 * self.split)
                if k == 'video':
                    pass
        logging.info(f"Picked {self.picked_streams['audio']} to use for condensing")
        logging.info(f"Picked {self.picked_streams['video']} to use for condensing")
        logging.info(f"Picked {self.picked_streams['subtitle']} to use for condensing")

    def process_subtitles(self):
        if self.picked_streams['subtitle'] is None:
            logging.error(f'No subtitle stream to process for output stem {self.outstem}')
            return
        if self.insufficient:
            return
        logging.debug("process_subtitles merged into choose_streams")
        # subfile = self.picked_streams['subtitle'].demux(overwrite_existing=self.demux_overwrite_existing)
        # times = subtools.load_subtitle_times(subfile.filepath)
        # times = subtools.merge_times(times, threshold=self.threshold, padding=self.padding)
        # self.dialogue_times = subtools.partition_and_split(sub_times=times, partition_size=1000*self.partition,
        #                                                    split_size=1000*self.split)

    def export_subtitles(self):
        if self.picked_streams['subtitle'] is None:
            logging.info(f'No subtitle stream to process for output stem {self.outstem}')
            return
        subpath = self.picked_streams['subtitle'].get_data_path()
        subext = subpath.suffix
        subdata = subtools.SubtitleManipulator(subpath, threshold=self.threshold, padding=self.padding)
        subdata.load(include_all=self.use_all_subs, regex=self.subtitle_regex_filter)
        subdata.merge_groups()
        subdata.condense()

        self.subtitle_outfile = Path(self.outdir) / (
                    self.outstem + f'.condensed{self.out_subext if self.out_subext is not None else subext}')
        subdata.condensed_ssadata.save(self.subtitle_outfile, encoding=u'utf-8')

    def export_audio(self):
        if self.picked_streams['audio'] is None:
            logging.error(f'No audio stream to process for output stem {self.outstem}')
            return
        outfile = Path(self.outdir) / (self.outstem + f'.{self.out_audioext}')
        # logging.info(f"exporting condensed audio to {outfile}")  # todo: fix output naming
        if outfile.exists() and not self.overwrite_existing_generated:
            logging.warning(f"Can't write to {outfile}: file exists and not set to overwrite")
            return
        export_condensed_audio(self.dialogue_times, audiofile=self.picked_streams['audio'].get_data_path(),
                               outfile=outfile)

    def export_video(self):
        if self.picked_streams['video'] is None:
            logging.error(f'No video stream to process for output stem {self.outstem}')
            return

        outfile = Path(self.outdir) / (self.outstem + '.mkv')
        logging.info(f"exporting condensed video to {outfile}")
        if outfile.exists() and not self.overwrite_existing_generated:
            logging.warning(f"Can't write to {outfile}: file exists and not set to overwrite")
            return
        export_condensed_video(self.dialogue_times, audiofile=self.picked_streams['audio'].get_data_path(),
                               subfile=self.subtitle_outfile,
                               videofile=self.picked_streams['video'].get_data_path(),
                               outfile=outfile)
        return

    def export(self):
        if self.insufficient:
            return
        subtools.get_compression_ratio(self.dialogue_times, self.picked_streams['audio'].demux_file.filepath)
        if self.condensed_subtitles:
            self.export_subtitles()
        if self.condensed_audio:
            self.export_audio()
        if self.condensed_video:
            self.export_video()

    def cleanup(self):
        if self.keep_temporaries:
            return
        for k in ['audio', 'video', 'subtitle']:
            if len(self.partitioned_streams) == 0:
                continue
            for s in self.partitioned_streams[k]:
                s.cleanup_demux()
